/* read a apple 2 disk image, extract the files, and convert all
   text files found to unix format */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>

typedef unsigned char byte;

typedef struct {
  byte bootblock [1024];
  byte x01 [6];
  byte dnlen;
  byte dname [7];
  byte blocks [2];
  byte nfiles [2];
  byte x02 [8];
  struct {
    byte firstblk [2];
    byte folloblk [2];
    byte filetype [2];
    byte dnlen;
    byte dname [15];
    byte lastused [2];
    byte xdate [2];
  } file [77];
} appledisk;

int g2b (byte *p)
{
  short r;
  memcpy (&r,p,2);
  return (r);
}

int g2s (byte *p, char *s)
{
  byte i, l, d;
  l = i = *p;
  d = 0;
  while (l > 0) {
    s[l] = d;
    d = p[l];
    if ((d < 33) || (d > 126) || (d == '/')) {
      d = '_';
    }
    l--;
  }
  *s = d;
  return (i);
}

void helper (char *n)
{
  fprintf (stderr,
  "Usage: %s <file>\n"
  "extracts from an apple 2 disk image all files and saves them in the current path\n",
  n);
  exit (1);
}

void ctt (FILE *f, byte *s)
{
  int i, l;
  byte e;
  const byte lf = 10;
  const byte bk = 32;
  l = 1024;
  while ((l > 0) && (!s[l-1])) {
    l--;
  }
  i = 0;
  while (i < l) {
    switch (s[i]) {
      case 13:
        if (fwrite (&lf,1,1,f) != 1) {
          fprintf (stderr,"write error, %s",strerror(errno));
          exit (1);
        }
        break;
      case 16:
        e = s[++i] - 32;
        while (e > 0) {
          if (fwrite (&bk,1,1,f) != 1) {
            fprintf (stderr,"write error, %s",strerror(errno));
            exit (1);
          }
          e--;
        }
        break;
      default:
        if (fwrite (&s[i],1,1,f) != 1) {
          fprintf (stderr,"write error, %s",strerror(errno));
          exit (1);
        }
        break;
    }
    i++;
  }
}

int main (int argc, char *argv[])
{
  int i, wf, wt, wy, wl, th, tx, ib;
  unsigned xd, xm, xy;
  FILE *fr;
  FILE *fw;
  FILE *ft;
  appledisk img;
  char diskname [519];
  char filename [257];
  byte b [1024];
  struct timeval tv[2];
  if (argc != 2) {
    fprintf (stderr,"Wrong number of arguments\n");
    helper (argv[0]);
  }
  if (!strcmp (argv[1],"--help")) {
    helper (argv[0]);
  }
  tx = 1;
  if ((fr = fopen (argv[1],"r")) == NULL) {
    fprintf (stderr,"cannot open %s, %s\n",argv[1],strerror(errno));
    exit (1);
  }
  if (fread (&img,sizeof(appledisk),1,fr) != 1) {
    fprintf (stderr,"read error, %s",strerror(errno));
    exit (1);
  }
  printf ("disk [%d] %s:, %d blocks, %d files\n",
    g2s(&img.dnlen,&diskname[0]),
    diskname,
    g2b(&img.blocks[0]),
    g2b(&img.nfiles[0]));
  if (mkdir (diskname,0777) < 0) {
    fprintf (stderr,"cannot open dir %s, errno=%s\n",diskname,strerror(errno));
    exit (1);
  }
  diskname[img.dnlen] = '/';
  i = 0;
  while (i < g2b(&img.nfiles[0])) {
    xd = g2b(&img.file[i].xdate[0]);
    xy = (xd >> 9) & 0x7f;
    xm = xd & 0x0f;
    xd = (xd >> 4) & 0x1f;
    printf ("%3d %3d %d %3d %4d/%2d/%2d [%2d] <%s>\n",
      wf = g2b(&img.file[i].firstblk[0]),
      wt = g2b(&img.file[i].folloblk[0]),
      wy = g2b(&img.file[i].filetype[0]),
      wl = g2b(&img.file[i].lastused[0]),
      xy + 1900, xm, xd,
      g2s(&img.file[i].dnlen,&filename[0]),
      filename);
    if (xm < 3) {
      xm += 12;
      xy -= 1;
    }
    tv[1].tv_sec = tv[0].tv_sec = ((xd + ((xm - 3) * 153 + 2) / 5
     + (xy * 1461) / 4 + ((xy / 100 + 1) * 3) / 4) - 25509) * 86400;
    tv[1].tv_usec = tv[0].tv_usec = 0;
    strcpy (&diskname[img.dnlen+1],filename);
    if ((fw = fopen (diskname,"w")) == NULL) {
      fprintf (stderr,"cannot open %s, %s\n",diskname,strerror(errno));
      exit (1);
    }
    if (wy == 3) {
      strcpy (&diskname[img.dnlen+1],"text");
      if (tx) {
        if (mkdir (diskname,0777) < 0) {
          fprintf (stderr,"cannot open dir %s, %s\n",diskname,strerror(errno));
          exit (1);
        }
        tx = 0;
      }
      strcpy (&diskname[strlen(diskname)],"/");
      strcpy (&diskname[strlen(diskname)],filename);
      if ((ft = fopen (diskname,"w")) == NULL) {
        fprintf (stderr,"cannot open %s, %s\n",diskname,strerror(errno));
        exit (1);
      }
      th = 2;
    }
    fseek (fr,512*wf,SEEK_SET);
    ib = 0;
    while (wf < wt) {
      if (fread (&b[ib],512,1,fr) != 1) {
        fprintf (stderr,"read error, %s",strerror(errno));
        exit (1);
      }
      wf++;
      if (fwrite (&b[ib],wf==wt?wl:512,1,fw) != 1) {
        fprintf (stderr,"write error, %s",strerror(errno));
        exit (1);
      }
      ib = 512-ib;
      if (wy == 3) {
        if (th > 0) {
          th--;
        } else {
          if (ib == 0) {
            ctt(ft,&b[0]);
          }
        }
      }
    }
    fclose (fw);
    if (wy == 3) {
      fclose (ft);
      utimes(&diskname[0], &tv[0]);
      strcpy(&diskname[img.dnlen+1],filename);
    }
    utimes(&diskname[0], &tv[0]);
    i++;
  }
}
